<?php
# $RCSfile: tnt-compiler.php $ $Revision: 1.2 $ $Date: 2009/09/19 01:25:00 $

# TNT: Trigger Node Template (for use with php 4, 5)
# Version 1.0 for php5 (26-Apr-2009).
# Copyright Malcolm Dew-Jones 2009
# License
# TNT is released under the LGPL (Lesser GPL). More info on the LGPL
# is available at http://www.gnu.org/copyleft/lesser.html

$tnt_previous_error_reporting = error_reporting(E_ALL);

class tnt_thing
{
    var $name;
    var $container;
    var $args;

    function tnt_thing($name,$args,&$container)
    {
        $this->args=$args;
        $this->name=$name;
        $this->container=&$container;
    }

    function name()       { return $this->name; }
    function &container() { return $this->container; }
    function args()       { return $this->args; }

    function top()
    {
        if ( isset($this->container) and
             method_exists($this->container, 'top')
           )
            return $this->container->top();
        else
            return $this;
    }

    function set_args($args) { $this->args = $args; }
}        

class tnt_literal
{
    var $text;
    var $container;
    var $encode;    # just boolean for now

    function tnt_literal($text,$encode,&$container)
    {
        $this->text   = $text;
        $this->encode = $encode;
##        $this->container=&$container;
    }
    
    function add_text($text)
    {
        $this->text .= $text;
    }

    function encode() { return $this->encode; }

    function as_code()
    {
            ## simple macro parsing could be improved
            ## unbalanced brackets not warned or error 
            ## we don't have a good way to pass the macro
            ## brackets in as options yet.

        $s1 =sprintf("%c?php",60);
        $s2 =sprintf("?%c",62);

        $bits = explode("{+",$this->text);

        if ($this->encode()) $code = htmlspecialchars(array_shift($bits));
        else                 $code =                  array_shift($bits);

        for($i=0;$i<count($bits);$i++)
        {
            list($macro,$string) = explode("+}",$bits[$i],2);
            if ($macro != '')    $macro  = "$s1 echo $macro;$s2";
            if (!isset($string)) $string = '';
            else
            if ( ## isset($string) and 
                 $string != '') 
            {
                if ($this->encode()) $string = htmlspecialchars($string);
            }
            $code .= $macro . $string;
        }

        return $code;
    }
}

class tnt_macro
{
    var $text;
    var $container;

    function tnt_macro($text,&$container)
    {
        $this->text = $text;
##        $this->container=&$container;
    }

    function as_code()
    {
        return "<?php echo $this->text; ?>";
    }
}

class tnt_call extends tnt_thing
{
    function tnt_call($name,$args,&$container)
    {
        $this->tnt_thing($name,$args,$container);
    }

    function as_code()
    {
        return '<?php '.$this->name() . '(' . $this->args() . ')' . "; ?>";
    }
}

class tnt_invoke extends tnt_thing
{
    var $lname;
    
    function tnt_invoke($name,$args,$lname,&$container)
    {
        $this->tnt_thing($name,$args,$container);
        $this->lname=$lname;
    }

    function logic_name() 
    {   
        if (!isset($this->lname) or $this->lname=='') return $this->name();
        else                                          return $this->lname;
    }

    function as_code()
    {
            ## the logic should be whatever the original definition of the 
            ## component being invoked would have been, but we don't look
            ## that up yet, the user must specify it.  Which also means
            ## it doesn't have to be accessible to the compiler
            ## i.e. it could be in another template entirely

        $name       = $this->name();
        $logic      = $this->logic_name() . '_logic';
        $more_args  = tnt_compiler::more_args($this->args());

        return "<?php \$this->tnt_logic_for_vobj->$logic( new $name(\$this) $more_args ); ?>";
    }
}

class tnt_visual extends tnt_thing
{
    var $runs;

    function tnt_visual($name,$args,&$container)
    {
        $this->tnt_thing($name,$args,$container);
        $this->runs   = array();
        $this->runs[] = & new tnt_literal('',false,$container);
    }

    function args()
    {
        $args = parent::args();
        if (isset($args)) return $args;
        else              return $this->container->args();
    }
    
    function add_text($text,$encode)
    {
        $run = & $this->runs[count($this->runs)-1];

        if ( get_class( $run ) == 'tnt_literal' and $run->encode() == $encode )
        {
            $run->add_text($text);
        }
        else
        {
            $this->runs[] = & new tnt_literal($text,$encode,$this->container);
        }
    }

    function add_macro($text)
    {
        $this->runs[] = & new tnt_macro($text,$this->container);
    }

    function add_call($func,$args)
    {
        $this->runs[] = & new tnt_call($func,$args,$this->container);
    }

    function add_invoke($invoke,$args,$logic)
    {
        $this->runs[] = & new tnt_invoke($invoke,$args,$logic,$this->container);
    }


    function add_component(&$component)
    {
        $this->runs[] = & $component;
    }
    
    function reset()
    {
        $this->runs   = array();
        $this->runs[] = & new tnt_literal('',false,$this->container);
    }

    function print_as_code()
    {
        $globals = $this->container->globals();
        if ($globals != '') $globals = "global $globals;\n";
        echo '  function ' , $this->name() ,'(', $this->args() , ")\n  {  $globals ?>";
        foreach ($this->runs as $run)
        {
            echo $run->as_code();
        }
        echo "<?php\n  }\n\n";
    }

}

class tnt_logic extends tnt_thing
{
    function tnt_logic($name,$args,&$container)
    {
        $this->tnt_thing($name,$args,$container);
    }

    function name()
    {
        $name = parent::name();
        if ($name) return $name;
        else       return $this->container->name();
    }

    function as_code()
    {
        $logic          = $this->name() . '_logic';
        $container_name = $this->container->name();
        $more_args      = tnt_compiler::more_args($this->args());

        return "<?php \$this->tnt_logic_for_vobj->$logic( new $container_name(\$this) $more_args); ?>";
    }
}


class tnt_top { } # just to allow a reference to it for first component

class tnt_component extends tnt_thing
{
    var $visuals;
    var $visual;
    var $logic;
    var $start;
    var $end;
    var $globals;
    var $vars=NULL;
    var $usage; # call invoke define apply

    function tnt_component($name,$args,&$container)
    {
        $this->tnt_thing($name,$args,$container);
        $this->start    = & new tnt_visual('start',NULL, $this);
        $this->visuals  = array();
        $this->end      = & new tnt_visual('end'  ,NULL, $this);
        $this->visual   = & $this->start;
        $this->logic    = new tnt_logic($name, NULL, $this);
        $this->usage    = 'apply';
    }

    function globals() { return $this->globals; }

    function set_globals($globals)
    { 
        $this->globals = $globals;
    }

    function vars() { return $this->vars; }

    function set_vars($vars)
    {
        $this->vars = $vars;
    }

    function as_code()
    {
        if     ($this->usage() == 'apply')  return $this->logic->as_code();
        elseif ($this->usage() == 'define') return '';
        else die('INTERNAL ERROR '.$this->name . ' bad usage:' . $this->usage());
    }

    function set_logic($name,$args)
    {
        $this->logic = & new tnt_logic($name, $args, $this);
    }

    function set_usage($usage) 
    { 
        $this->usage = $usage; 
    }

    function usage() { return $this->usage ; }

    function add_text($text,$encode)
    {
        $this->visual->add_text($text,$encode);
    }

    function add_macro($text)
    {
        $this->visual->add_macro($text);
    }

    function add_invoke($invoke, $args, $logic)
    {
        $this->visual->add_invoke($invoke, $args, $logic);
    }

    function add_call($func,$args)
    {
        $this->visual->add_call($func,$args);
    }

    function start_child($name,$args)
    {   
        $this->visuals[] = & new tnt_visual($name, $args, $this);
        $this->visual    = & $this->visuals[count($this->visuals)-1];
    }

    function end_child()
    {
        $this->visual = & $this->end;
        $this->end->reset();
    }

    function add_component(&$component)
    {
        $this->visual->add_component($component);
    }

    function print_as_code()
    {

        if ( get_class( $this->container ) == 'tnt_top' )
        {
            $var_defs = ""
            ."  var \$tnt_container_vobj;\n"
            ."  var \$tnt_logic_for_vobj;\n";

            $get_var_def = ""
            ."  function tnt_get_var(\$var_name,\$default='')\n"
            ."  {\n"
            ."      if ( method_exists(\$this, \$var_name))\n"
            ."          return \$this->\$var_name();\n"
            ."  \n"
            ."      if ( isset(\$this->tnt_container_vobj) and\n"
            ."           method_exists(\$this->tnt_container_vobj, 'tnt_get_var')\n"
            ."         )\n"
            ."          return \$this->tnt_container_vobj->tnt_get_var(\$var_name,\$default);\n"
            ."      return \$default;\n"
            ."  }\n\n"
            ."  function tnt_logic()\n"
            ."  {\n"
            ."      return \$this->tnt_logic_for_vobj;\n"
            ."  }\n\n"
            ."  function tnt_set_logic(\$tnt_logic_for_vobj)\n"
            ."  {\n"
            ."      \$this->tnt_logic_for_vobj=\$tnt_logic_for_vobj;\n"
            ."      return \$this->tnt_logic_for_vobj;\n"
            ."  }\n\n"
            ."  function tnt_init_this_using(\$container)\n"
            ."  {\n"
            ."      if (isset(\$container->tnt_logic_for_vobj))\n"
            ."      {\n"
            ."          \$this->tnt_container_vobj=\$container;\n"
            ."          \$this->tnt_logic_for_vobj=\$container->tnt_logic_for_vobj;\n"
            ."      }else\n"
            ."      {\n"
            ."          \$this->tnt_logic_for_vobj=\$container;\n"
            ."      }\n"
            ."  }\n\n"
            ."";

            $extends = "";
        }else
        {
            $var_defs = "";
            $top=$this->top();
            $extends = "extends ".$top->name();
            $get_var_def = '';
        }

        $vars     = $this->vars();
        $get_sets = '';

        if (isset($vars))
        foreach ($this->vars as $var)
        {
            $var_defs .= "  var $var;\n";
            list($var_name) = explode('=',$var);
            $var_name = substr($var_name,1); # get rid of the $
            
            $get_sets .= 
              "  function $var_name() { return \$this->$var_name; }\n"
            . "  function set_$var_name(\$$var_name) { \$this->$var_name=\$$var_name; }\n";
        }

        echo "\n\nclass ".$this->name(). " $extends\n{ \n$var_defs$get_sets";

        echo "\n";
        echo "  function $this->name(\$container=NULL)\n";
        echo "  {\n";
        echo "      \$this->tnt_init_this_using(\$container);\n";
        echo "  }\n\n";

        echo $get_var_def;

        $this->start->print_as_code();
        foreach ($this->visuals as $visual)
        {
            $visual->print_as_code();
        }
        $this->end->print_as_code();
        echo "} # end class ".$this->name()."\n";
    }

}


class tnt_compiler
{
    var $current;
    var $top;
    var $parser;
    var $recent_tagname;
    var $depth;
    var $unnamed=1;
    var $logic_code='';
    var $prefix='tnt';
    var $prefix_len=4; # includes :

    #KLUDGE options
    var $remove_nonhtml_end_tags=true;
    var $kludge_nbsp_translate=true; ## false;
    var $kludge_br_etc_tags_translate=false;
    var $add_xml_PI_line1=false;
    var $case_folding=false;

    #Other options
    var $reencode_chardata=false;
    var $lint_warnings=false;
    var $apply_or_define='apply';

function set_lint_warnings($on_or_off)
{
    $this->lint_warnings=$on_or_off;
}

function set_kludge_nbsp_translate($on_or_off)
{
    $this->kludge_nbsp_translate = $on_or_off;
}

function set_kludge_br_etc_tags_translate($on_or_off)
{
    $this->kludge_br_etc_tags_translate = $on_or_off;
}

function set_reencode_chardata($on_or_off)
{
    $this->reencode_chardata = $on_or_off;
}

function set_remove_nonhtml_end_tags($on_or_off)
{
    $this->remove_nonhtml_end_tags = $on_or_off;
}

function set_apply_or_define($apply_or_define)
{
    $this->apply_or_define = $apply_or_define;
}

function set_tnt_prefix($prefix)
{
    $this->prefix = $prefix;
    $this->prefix_len = strlen($prefix)+1;
}


function add_logic_code($more_code)
{
    $this->logic_code = $this->logic_code . $more_code;
}

# why not the php function, was there a reason?
function attrib_escape( $s )
{
    $s = str_replace ("&" , "&amp;" , $s );
    $s = str_replace ("<" , "&lt;" , $s );
    $s = str_replace (">" , "&gt;" , $s );
    $s = str_replace ('"' , "&quot;" , $s );
    return $s;
}

function chardata_escape( $s )
{
    return $this->attrib_escape($s);
}

function gen_name($n)
{
    return "tnt_noname_${n}_" . $this->unnamed++;
}

function attributes_to_xml($attributes)
{
    $xml = "";
    $tnt     = $this->prefix;
    $tnt_len = $this->prefix_len;

    foreach ($attributes as $attribute_key => $attribute_value) 
    {
        # we never copy tnt: attribs
        if (substr($attribute_key,0,$tnt_len) != "$tnt:")
        {
            $value = $this->attrib_escape($attribute_value);
            $xml .= sprintf(' %s="%s"', $attribute_key, $value);
        }
    }
    return $xml;
}

function start_element_to_xml($name, $attrs)
{
    return "<$name" . $this->attributes_to_xml($attrs) . ">";
}

function end_element_to_xml($name)
{
    # no end tag in html for some tags
    if ( !$this->remove_nonhtml_end_tags or
         strcasecmp($name,'br') and
         strcasecmp($name,'hr') and
         strcasecmp($name,'input') and
         1==1
       ) 
       return "</$name>";
}

function more_args($args)
{
    if (!isset($args) or $args=='')  return '';
    else                             return ',' . $args;
}


function foundat()
{
    return "line " . xml_get_current_line_number($this->parser)
           . ", tag $this->recent_tagname, (tag starts at column " 
           . (xml_get_current_column_number($this->parser)+1)
           . ")";
}

function tnt_throw($msg)
{
    ob_end_clean();
    $msg = "Template format error: $msg, at ".$this->foundat()."\n";
    die($msg);
}

function warn($msg)
{
    $msg = "\necho 'Template format warning: $msg, at ".$this->foundat()."\n';\n";
    echo $msg;
}

function parse_vars($list)
{
    return explode(',',$list);
}

function parse_thing($name)
{
    if (!isset($name)) return array('',NULL);

    $args = strstr($name,'(');
    if ($args)
    {   
        $func = trim(substr($name,0,strpos($name,'(' )));
        # remove brackets
        $args=trim($args);
        if (substr($args,-1) != ')' ) 
        {
            $this->tnt_throw("args didn't end with ')' in '$name'");
        }
        $args = trim(substr($args,1,strlen($args)-2));
        $name=$func;
    }else
    {
        $args=NULL;
    }

    return array(trim($name),$args);
}        

function startElement($parser, $tagname, $attrs)
{
    $this->recent_tagname = $tagname;
    array_unshift( $this->depth , '' );
    $tnt     = $this->prefix;
    $tnt_len = $this->prefix_len;

    # attribute tnt:name    ="name(args)"
    # attribute tnt:logic   ="name(args)"
    # attribute tnt:globals ="vars"
    # attribute tnt:vars    ="vars"
    # attribute tnt:child   ="name(args)"
    # attribute tnt:apply   ="true" "false"

    if (substr($tagname,0,$tnt_len) =="$tnt:") 
    {   
        if (!isset($attrs["$tnt:name"   ]) and isset($attrs['name'   ])) $attrs["$tnt:name"   ] = $attrs['name'];
        if (!isset($attrs["$tnt:logic"  ]) and isset($attrs['logic'  ])) $attrs["$tnt:logic"  ] = $attrs['logic'];
        if (!isset($attrs["$tnt:globals"]) and isset($attrs['globals'])) $attrs["$tnt:globals"] = $attrs['globals'];
        if (!isset($attrs["$tnt:vars"   ]) and isset($attrs['vars'   ])) $attrs["$tnt:vars"   ] = $attrs['vars'];
        if (!isset($attrs["$tnt:child"  ]) and isset($attrs['child'  ])) $attrs["$tnt:child"  ] = $attrs['child'];
        if (!isset($attrs["$tnt:apply"  ]) and isset($attrs['apply'  ])) $attrs["$tnt:apply"  ] = $attrs['apply'];
        if (!isset($attrs["$tnt:call"   ]) and isset($attrs['call'   ])) $attrs["$tnt:call"   ] = $attrs['call'];
        if (!isset($attrs["$tnt:invoke" ]) and isset($attrs['invoke' ])) $attrs["$tnt:invoke" ] = $attrs['invoke'];

        $keep=false;
    }
    else
    {   
        $keep=true;

        if ($this->lint_warnings)
        {   
            $suspicious = '';
            ## name is used so often it doesn't help to check it
            ## we may need another name for name (e.g. part)
            ## if (isset($attrs['name'   ])) $suspicious .= ", name";

            if (isset($attrs['logic'  ])) $suspicious .= ", logic";
            if (isset($attrs['globals'])) $suspicious .= ", globals";
            if (isset($attrs['vars'   ])) $suspicious .= ", vars";
            if (isset($attrs['child'  ])) $suspicious .= ", child";
            if (isset($attrs['apply'  ])) $suspicious .= ", apply";
            if (isset($attrs['call'   ])) $suspicious .= ", call";
            if (isset($attrs['invoke' ])) $suspicious .= ", invoke";
                
            if ($suspicious != '')
            $this->warn("suspicious attribute(s)$suspicious, may be missing tnt: prefix");
        }

    }

    if (isset($attrs["$tnt:invoke"]))
    {
        list($invoke,$args) = $this->parse_thing($attrs["$tnt:invoke"]);
        if ($invoke == '') $this->tnt_throw("invoke value is empty");

        if (isset($attrs["$tnt:logic"]) and trim($attrs["$tnt:logic"]) == '')
            $this->tnt_throw("logic value is empty");
        if (!isset($attrs["$tnt:logic"])) $attrs["$tnt:logic"]=''; # avoid unset error messages
        list($logic,$largs) = $this->parse_thing($attrs["$tnt:logic"]);

        if (isset($args) and isset($largs) ) $this->tnt_throw("at most one set of args can be specified");
        if (!isset($args)) $args = $largs;
    
        $this->current->add_invoke( $invoke, $args, $logic);
    }

    elseif (isset($attrs["$tnt:name"]) or isset($attrs["$tnt:logic"]) )
    {
        if (isset($attrs["$tnt:name"]) and trim($attrs["$tnt:name"]) == '')
            $this->tnt_throw("name value is empty");
                   
        if (isset($attrs["$tnt:logic"]) and trim($attrs["$tnt:logic"]) == '')
            $this->tnt_throw("logic value is empty");

        if (!isset($attrs["$tnt:name" ])) $attrs["$tnt:name" ]=''; # avoid unset error messages
        if (!isset($attrs["$tnt:logic"])) $attrs["$tnt:logic"]=''; # avoid unset error messages

        list($part,$args)   = $this->parse_thing($attrs["$tnt:name"]);
        list($logic,$largs) = $this->parse_thing($attrs["$tnt:logic"]);

        if ($part == '' and $logic == '') $this->tnt_throw("at least one of name or logic must have a NAME");
        if ($part == '') $part = $this->gen_name($logic);

        $this->depth[0]='part';

        $previous = & $this->current;
        $this->current = & new tnt_component($part , $args, $previous);

        if ($attrs["$tnt:logic"] != '') $this->current->set_logic($logic, $largs);

        if (isset($attrs["$tnt:apply"]))
        {
            if (strcasecmp($attrs["$tnt:apply"],'false')==0)
            {
                $this->current->set_usage('define');
            }            
            elseif (strcasecmp($attrs["$tnt:apply"],'true')==0)
            {
                $this->current->set_usage('apply');
            }
            else
            {   $v=$attrs["$tnt:apply"];
                $this->tnt_throw("apply value must be 'true' or 'false', not '$v'");
            }
        }
        else
        {   $this->current->set_usage($this->apply_or_define);
        }

        if (isset($attrs["$tnt:globals"]))
        {   
            $this->current->set_globals( $attrs["$tnt:globals"] );
        }

        if (isset($attrs["$tnt:vars"]))
        {
            if (trim($attrs["$tnt:vars"]) == '') $this->tnt_throw("vars value is empty");
            $vars = $this->parse_vars($attrs["$tnt:vars"]);
            $this->current->set_vars($vars);
        }
        
        $previous->add_component($this->current);
    }
    
    elseif (isset($attrs["$tnt:child"]))
    {
        list($child,$args) = $this->parse_thing($attrs["$tnt:child"]);
        if ($child == '') $this->tnt_throw("child value is empty");

        if (isset($attrs["$tnt:globals"]))
        {   
            $prev_globals = $this->current->globals();
            if (isset($prev_globals)) $this->tnt_throw("globals already defined for this component");
            $this->current->set_globals( $attrs["$tnt:globals"] );
        }

        # note: vars cannot be set like this because if they are set in the template 
        # top it breaks the logic of looking them up.
        
        $this->depth[0]='child';

        $this->current->start_child($child, $args);
    }
    
    elseif (isset($attrs["$tnt:call"]))
    {
        list($call,$args) = $this->parse_thing($attrs["$tnt:call"]);
        if ($call == '') $this->tnt_throw("call value is empty");
        $this->current->add_call( $call, $args);
    }
    
    if ($keep)  $this->current->add_text( $this->start_element_to_xml($tagname, $attrs) , false );
}

function endElement($parser, $tagname)
{
    $this->recent_tagname = $tagname;
    $part = array_shift( $this->depth );
    $tnt     = $this->prefix;
    $tnt_len = $this->prefix_len;

    if (substr($tagname,0,$tnt_len) != "$tnt:")
        $this->current->add_text( $this->end_element_to_xml($tagname) , false );

    if      ($part == 'child') 
    {   
        $this->current->end_child();
    }

    elseif  ($part == 'part')  
    {
        $this->current->print_as_code();
        $this->current =& $this->current->container();
    }
}

function characterData($parser, $data)
{
    $this->current->add_text( $data , $this->reencode_chardata );
}

function otherData($parser, $data)
{
    $this->current->add_text( $data , false );
}

function processingInstruction($parser, $target, $data)
{
    $tnt = $this->prefix;

    if (strcasecmp($target,"$tnt") ==0)
    {
        if (isset( $data ) and $data != '') eval( $data );
    }

    elseif (strcasecmp($target,"$tnt:code") ==0)
    {
        if (isset( $data ) and $data != '') $this->add_logic_code( $data );
    }

    elseif (strcasecmp($target,'php') ==0)
    {
        $s1 =sprintf("%c?",60);
        $s2 =sprintf("?%c",62);

        $this->current->add_text( "$s1$target " , false);
        $this->current->add_text( $data         , false);
        $this->current->add_text( $s2           , false);
    }

    else
    {
        $this->current->add_macro( "sprintf('%c?$target ',60)");
        $this->current->add_text( $data, false);
        $this->current->add_macro( "sprintf('?%c',62)");
    }
}

function tnt_compiler()
{
    $this->parser = xml_parser_create();
    xml_set_object($this->parser, $this);

    xml_parser_set_option ($this->parser, XML_OPTION_CASE_FOLDING, $this->case_folding);

    xml_set_element_handler($this->parser, "startElement", "endElement");
    xml_set_character_data_handler($this->parser, "characterData");
    xml_set_processing_instruction_handler($this->parser, "processingInstruction");
    xml_set_default_handler($this->parser, "otherData");

}

function load_from_file($file)
{

    if (!($fp = fopen($file, "r"))) 
    {
        die("Could not open $file");
    }

    $data = fread($fp, 4096);
    $data = ltrim($data);

    if ($this->add_xml_PI_line1 and substr($data,0,5) != "<?xml")
        $data = '<?xml version="1.0" ?>' . $data;

    # HTML NBSP KLUDGE, not so kludgey as last time, but still pretty dramatic
    if ($this->kludge_nbsp_translate)
    {
        if (substr($data,0,5) != "<?xml")
        {   $data = $this->tnt_added_HTML_ENTITIES() . $data;
        }else
        {
            $gt_at = strpos($data,'>');
            $left  = substr($data,0,$gt_at+1);
            $right = substr($data,$gt_at+1);
            $data = $left . $this->tnt_added_HTML_ENTITIES() . $right;
        }

    }

    list($name) = split('\.[a-zA-Z]*$',basename($file));
    $this->top =& new tnt_component($name,'',new tnt_top);
    $this->current =& $this->top;
    $this->depth = array();

    while ($data != '')
    {
        ## HTML BR to BR/ HR to HR/ KLUDGE
        if ($this->kludge_br_etc_tags_translate)
        {
            $data = str_replace("<br>" , "<br/>" , $data);
            $data = str_replace("<hr>" , "<hr/>" , $data);
        }

        if (!xml_parse($this->parser, $data, feof($fp))) 
        {
            $err = 
             sprintf( "XML error: %s, in %s at line %d, column %d\n"
                    , xml_error_string(xml_get_error_code($this->parser))
                    , $file
                    , xml_get_current_line_number($this->parser)
                    , xml_get_current_column_number($this->parser)+1
                    );
            ob_end_clean();
            die($err);
        }
        $data = fread($fp, 4096);
    }
    xml_parser_free($this->parser);

    if ( isset( $this->logic_code ) )
    {   
        echo $this->logic_code ;
    }

    $this->top->print_as_code();
    return $name;
}

function tnt_added_HTML_ENTITIES()
{
    $dtd = "<!DOCTYPE tnt_added_HTML_ENTITIES [\n";
    $trans = get_html_translation_table(HTML_ENTITIES);
    foreach ($trans as $entity) 
    {   $ntit = substr($entity,1,-1);
        $dtd .= "<!ENTITY $ntit \"$ntit\">\n";
    }   
    $dtd .= "]>\n";
    return $dtd;
}

} # END TNT_COMPILER CLASS

# TEST OPTION
if (!empty($argv))
if (!empty($argv[1]))
{
    $file = $argv[1];
    $s1 =sprintf("%c?php",60);
    $s2 =sprintf("?%c",62);
    echo "$s1\n";
    $tnt=new tnt_compiler();
    $tnt->load_from_file($file);
    echo "$s2\n";
}

error_reporting($tnt_previous_error_reporting);

?>